#!/usr/bin/env python3
"""
## Network and Disk I/O Stats Monitor.

Monitor and log Network and Disk stats during a PyTorch Lightning training.

Copyright 2021 The PyTorch Lightning Team under Apache 2.0 <http://www.apache.org/licenses/LICENSE-2.0>
Copyright 2021 Leo Gallucci               under Apache 2.0 <http://www.apache.org/licenses/LICENSE-2.0>
"""
from typing import Any
from typing import Dict
from typing import Optional

import pytorch_lightning as pl
from pytorch_lightning.callbacks.base import Callback
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.utilities.exceptions import MisconfigurationException
from pytorch_lightning.utilities.parsing import AttributeDict
from pytorch_lightning.utilities.types import STEP_OUTPUT

from iometrics import DiskMetrics
from iometrics import NetworkMetrics


# Metrics generated:
LOG_KEY_NETW_BYTES_RECV = "network/recv_MB_per_sec"
LOG_KEY_NETW_BYTES_SENT = "network/sent_MB_per_sec"
LOG_KEY_DISK_UTIL = "disk/util%"
LOG_KEY_DISK_MB_READ = "disk/read_MB_per_sec"
LOG_KEY_DISK_MB_WRIT = "disk/writ_MB_per_sec"
LOG_KEY_DISK_IO_READ = "disk/io_read_count_per_sec"
LOG_KEY_DISK_IO_WRIT = "disk/io_writ_count_per_sec"
LOG_KEY_DISK_IO_WAIT = "disk/io_wait%"


class NetworkAndDiskStatsMonitor(Callback):

    r"""
    Automatically monitors and logs Network and Disk stats during training stage.

    ``NetworkAndDiskStatsMonitor`` is a callback and in order to use it you need to assign a logger in the ``Trainer``.

    Args:
        track_network_utilization: Set to ``True`` to monitor network bytes send and received.
            at the start and end of each step. Default: ``True``.
        track_disk_utilization: Set to ``True`` to monitor Disk read, write, IO/s and percentage of Disk utilization.
            at the start and end of each step. Default: ``True``.

    Example::

        >>> from pytorch_lightning import Trainer
        >>> from iometrics.pytorch_lightning.callbacks import NetworkAndDiskStatsMonitor
        >>> net_disk_stats = NetworkAndDiskStatsMonitor()
        >>> trainer = Trainer(callbacks=[net_disk_stats]) # doctest: +SKIP

    Metrics generated:

    - **LOG_KEY_NETW_BYTES_RECV** – Received MegaBytes per second (MB/s) on all relevant network interfaces as a SUM.
    - **LOG_KEY_NETW_BYTES_SENT** – Sent     MegaBytes per second (MB/s) on all relevant network interfaces as a SUM.
    - **LOG_KEY_DISK_UTIL**       – Disk utilization percentage as the average of all disk devices.
    - **LOG_KEY_DISK_MB_READ**    – Disks read MB/s    as the sum of all disk devices.
    - **LOG_KEY_DISK_MB_WRIT**    – Disks written MB/s as the sum of all disk devices.
    - **LOG_KEY_DISK_IO_READ**    – Disks read I/O operations per second    as the sum of all disk devices.
    - **LOG_KEY_DISK_IO_WRIT**    – Disks written I/O operations per second as the sum of all disk devices.
    - **LOG_KEY_DISK_IO_WAIT**    – Disks I/O percentage of time that the CPU is waiting.

    Raises
    ------
     MisconfigurationException
        When something prevents this utility from being set up.

    """

    def __init__(
        self,
        track_network_utilization: bool = True,
        track_disk_utilization: bool = True,
    ):
        super().__init__()

        # AttributeDict is a subclass of dict that allows to access keys as attributes using dot notation.
        self._settings = AttributeDict(
            {
                "track_network_utilization": track_network_utilization,
                "track_disk_utilization": track_disk_utilization,
            }
        )

    def setup(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", stage: Optional[str] = None) -> None:
        if not trainer.logger:
            raise MisconfigurationException(
                "Cannot use NetworkAndDiskStatsMonitor callback with Trainer that has no logger."
            )

    def on_train_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._net_meter: Any[NetworkMetrics, None] = None
        self._disk_meter: Any[DiskMetrics, None] = None

    def _get_new_logs(self) -> Dict[str, float]:
        new_logs: Dict[str, float] = {}

        if self._settings.track_network_utilization:
            if not hasattr(self, "_net_meter") or self._net_meter is None:
                self._net_meter = NetworkMetrics()
            self._net_meter.update_stats()
            new_logs[LOG_KEY_NETW_BYTES_RECV] = float(self._net_meter.mb_recv_ps.val)
            new_logs[LOG_KEY_NETW_BYTES_SENT] = float(self._net_meter.mb_sent_ps.val)

        if self._settings.track_disk_utilization:
            if not hasattr(self, "_disk_meter") or self._disk_meter is None:
                self._disk_meter = DiskMetrics()
            self._disk_meter.update_stats()
            new_logs[LOG_KEY_DISK_UTIL] = float(self._disk_meter.io_util.val)
            new_logs[LOG_KEY_DISK_MB_READ] = float(self._disk_meter.mb_read.val)
            new_logs[LOG_KEY_DISK_MB_WRIT] = float(self._disk_meter.mb_writ.val)
            new_logs[LOG_KEY_DISK_IO_READ] = float(self._disk_meter.io_read.val)
            new_logs[LOG_KEY_DISK_IO_WRIT] = float(self._disk_meter.io_writ.val)
            new_logs[LOG_KEY_DISK_IO_WAIT] = float(self._disk_meter.io_wait.val)

        return new_logs

    @rank_zero_only
    def on_train_batch_start(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        if not self._should_log(trainer):
            return

        logs: Dict[str, float] = self._get_new_logs()

        trainer.logger.log_metrics(logs, step=trainer.global_step)

    @rank_zero_only
    def on_train_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        if not self._should_log(trainer):
            return

        logs: Dict[str, float] = self._get_new_logs()

        trainer.logger.log_metrics(logs, step=trainer.global_step)

    @staticmethod
    def _should_log(trainer: "pl.Trainer") -> bool:
        modulo_result: bool = (trainer.global_step + 1) % trainer.log_every_n_steps == 0

        return modulo_result or trainer.should_stop


# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "NetworkAndDiskStatsMonitor",
    "LOG_KEY_NETW_BYTES_RECV",
    "LOG_KEY_NETW_BYTES_SENT",
    "LOG_KEY_DISK_UTIL",
    "LOG_KEY_DISK_MB_READ",
    "LOG_KEY_DISK_MB_WRIT",
    "LOG_KEY_DISK_IO_READ",
    "LOG_KEY_DISK_IO_WRIT",
]



#!/usr/bin/env python3
"""
## Classes to store Network I/O Statistics.

Ground truth comes from `/proc/net/dev`, a file with stats updated by the *nix kernel.

:copyright: (c) 2018 by Yaroslav Bulatov under The Unlicense <https://unlicense.org>
:copyright: (c) 2021 by Leo Gallucci under Apache License 2.0.
:license: Apache 2.0, see LICENSE for more details.
"""
import re
import time
from dataclasses import dataclass
from typing import Dict
from typing import List

from iometrics.average_metrics import AverageMetrics


@dataclass
class NetworkStats:

    """Simple data class to store network's relevant statistics."""

    bytes_recv: int = 0
    bytes_sent: int = 0


class NetworkMetrics:

    """Tracks and computes network received and sent MBytes/s metrics."""

    # pylint: disable=too-few-public-methods
    # One is reasonable in this case.

    def __init__(self) -> None:
        self.mb_recv_ps = AverageMetrics()
        self.mb_sent_ps = AverageMetrics()

        self.last_stats: Dict[str, NetworkStats] = get_network_bytes()
        self.last_log_time: float = time.time()

    def update_stats(self) -> None:
        """Compute metrics since last measurement then returns stats per second."""
        time_delta: float = time.time() - self.last_log_time

        # Important: You should wait at least 1 second between calls to the get_network_bytes function.
        # Before ~0.9 seconds `/proc/net/dev` will show the exact same values as last the time.
        if time_delta < 0.99:
            return

        new_stats: Dict[str, NetworkStats] = get_network_bytes()

        aggr_mb_recv_ps: float = 0.0
        aggr_mb_sent_ps: float = 0.0

        for device_name, last_stats in self.last_stats.items():
            bytes_recv_delta: int = new_stats[device_name].bytes_recv - last_stats.bytes_recv
            bytes_sent_delta: int = new_stats[device_name].bytes_sent - last_stats.bytes_sent

            # There's a bug that sometimes the delta is negative messing up the average.
            bytes_recv_delta = max(0, bytes_recv_delta)
            bytes_sent_delta = max(0, bytes_sent_delta)

            mb_recv_ps: float = bytes_recv_delta / 1e6 / time_delta
            mb_sent_ps: float = bytes_sent_delta / 1e6 / time_delta

            aggr_mb_recv_ps += mb_recv_ps
            aggr_mb_sent_ps += mb_sent_ps

        self.mb_recv_ps.update(aggr_mb_recv_ps)
        self.mb_sent_ps.update(aggr_mb_sent_ps)

        self.last_log_time = time.time()
        self.last_stats = new_stats


def get_network_bytes() -> Dict[str, NetworkStats]:
    """Return received, transmitted bytes (since the kernel started)."""
    # Important: You should wait at least 1 second between calls to this function.
    # Before ~0.9 seconds `/proc/net/dev` will show the exact same values as last the time.

    interface_filter_out = re.compile(r"^(lo|tun.+|face.+|bond.+|.+\.\d+)$")

    # Note: all counters at /proc/* are starting with zero when the kernel starts.
    with open("/proc/net/dev") as file:
        content: str = file.read()

    lines: List[str] = content.splitlines()
    lines = lines[2:]  # strip header

    # Accumulate the stats of each device:
    stats: Dict[str, NetworkStats] = {}

    for device_line in lines:
        fields: List[str] = device_line.strip().split()
        device_name: str = fields[0].strip(":")

        if interface_filter_out.match(device_name):
            continue

        device_stats = NetworkStats(
            # Using max because sometimes it returns a negative (wrong) number
            bytes_recv=max(0, int(fields[1])),
            bytes_sent=max(0, int(fields[9])),
        )

        stats[device_name] = device_stats

    return stats


# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "NetworkMetrics",
]



#!/usr/bin/env python3
"""
## Class to store Generic Statistics.

Computes and stores the average and current value of some metric.

:copyright: (c) 2018 by Yaroslav Bulatov under The Unlicense <https://unlicense.org>
:copyright: (c) 2021 by Leo Gallucci under Apache License 2.0.
:license: Apache 2.0, see LICENSE for more details.
"""
from dataclasses import dataclass


@dataclass
class AverageMetrics:

    """Tracks and computes generic metrics along with average, count, etc."""

    avg_mom: float = 0.5
    val: float = 0.0
    avg: float = 0.0
    smooth_avg: float = 0.0
    tot_sum: float = 0.0
    count: int = 0

    def __init__(self, avg_mom: float = 0.5) -> None:
        self.avg_mom = avg_mom
        self.reset()

    def reset(self) -> None:
        """Reset values when starting a new epoch (for example)."""
        self.val = 0.0
        self.avg = 0.0
        self.smooth_avg = 0.0
        self.tot_sum = 0.0
        self.count = 0

    def update(self, val: float) -> None:
        """Update last value and compute average, count, etc."""
        # Use max(0) to prevent rounding -0.0 negative numbers
        self.val = max(0, val)
        self.tot_sum += val
        if self.count == 0:
            self.smooth_avg = val
        else:
            self.smooth_avg = self.avg * self.avg_mom + val * (1 - self.avg_mom)
        self.count += 1
        self.avg = self.tot_sum / self.count


# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "AverageMetrics",
]



#!/usr/bin/env python3
"""
## Classes to store Disk I/O Statistics.

Ground truth comes from `/proc/diskstats`, a file with stats updated by the *nix kernel.

The list of devices comes from `/sys/block` directory.

:copyright: (c) 2021 by Leo Gallucci.
:license: Apache 2.0, see LICENSE for more details.
"""
import os
import time
from dataclasses import dataclass
from typing import Dict
from typing import List

import psutil

from iometrics.average_metrics import AverageMetrics


@dataclass
class DiskStats:

    """Simple data class to store disks relevant statistics."""

    kb_read: int = 0
    kb_writ: int = 0
    io_read: int = 0
    io_writ: int = 0
    io_util: int = 0


@dataclass
class AggregateDiskStats:

    """Simple data class to store disks aggregate relevant statistics."""

    mb_read_ps: float = 0.0
    mb_writ_ps: float = 0.0
    io_read_ps: float = 0.0
    io_writ_ps: float = 0.0
    io_util: float = 0.0


class DiskMetrics:

    """Tracks and computes disks read/written MBytes/s, also utilization and io counts metrics."""

    def __init__(self) -> None:
        self.mb_read = AverageMetrics()
        self.mb_writ = AverageMetrics()
        self.io_read = AverageMetrics()
        self.io_writ = AverageMetrics()
        self.io_util = AverageMetrics()
        self.io_wait = AverageMetrics()

        self.non_virtual_devices = get_non_virtual_disk_devices()

        self.last_stats: Dict[str, DiskStats] = self.get_disks_stats()
        self.last_log_time: float = time.time()

    def get_disks_stats(self) -> Dict[str, DiskStats]:
        """Return number of disk reads, writes, io (since the kernel started)."""
        # Note: all counters at /proc/* are starting with zero when the kernel starts.
        with open("/proc/diskstats") as file:
            content: str = file.read()

        lines = content.splitlines()

        # Accumulate the stats of each device:
        stats: Dict[str, DiskStats] = {}

        for device_line in lines:
            fields: List[str] = device_line.strip().split()
            device_name: str = fields[2]

            if device_name not in self.non_virtual_devices:
                continue

            device_stats = DiskStats(
                io_read=int(fields[3]),
                io_writ=int(fields[7]),
                io_util=int(fields[12]),
                kb_read=int(fields[5]),
                kb_writ=int(fields[9]),
            )

            stats[device_name] = device_stats

        return stats

    def update_stats(self) -> None:
        """Compute metrics since last measurement then returns stats per second."""
        time_delta: float = time.time() - self.last_log_time

        per_device_stats: Dict[str, DiskStats] = self.get_disks_stats()

        # Disks I/O percentage of time that the CPU is waiting
        avg_io_wait_since_last_read: float = psutil.cpu_times_percent(interval=0).iowait

        aggr = AggregateDiskStats()

        for device_name, last_all_devices_stats in self.last_stats.items():
            new_aggr = compute_new_stats_ps(per_device_stats, last_all_devices_stats, device_name, time_delta)

            aggr.mb_read_ps += new_aggr.mb_read_ps
            aggr.mb_writ_ps += new_aggr.mb_writ_ps
            aggr.io_read_ps += new_aggr.io_read_ps
            aggr.io_writ_ps += new_aggr.io_writ_ps
            aggr.io_util += new_aggr.io_util

        avg_disks_io_util: float = aggr.io_util / len(self.last_stats)

        self.mb_read.update(aggr.mb_read_ps)
        self.mb_writ.update(aggr.mb_writ_ps)
        self.io_read.update(aggr.io_read_ps)
        self.io_writ.update(aggr.io_writ_ps)
        self.io_util.update(avg_disks_io_util)
        self.io_wait.update(avg_io_wait_since_last_read)

        self.last_log_time = time.time()
        self.last_stats = per_device_stats


def get_non_virtual_disk_devices() -> List[str]:
    """Return a list of currently attached Disk names to which makes sense to measure I/O performance."""
    disks_devices_dir = "/sys/block"

    non_virtual_devices: List[str] = []

    for device_name in os.listdir(disks_devices_dir):
        abs_device_path = os.path.join(disks_devices_dir, device_name)

        if os.path.islink(abs_device_path):
            link_target = os.readlink(abs_device_path)
            if "virtual" not in link_target:
                non_virtual_devices.append(device_name)

    return non_virtual_devices


def compute_new_stats_ps(
    per_device_stats: Dict[str, DiskStats], last_all_devices_stats: DiskStats, device_name: str, time_delta: float
) -> AggregateDiskStats:
    """Compute the new stats per second based on last ones and the time delta."""
    aggr = AggregateDiskStats()

    bytes_read_delta: int = 0
    bytes_writ_delta: int = 0

    # Devices like AWS EBS or USB drives can get dynamically detached so validate key:
    if device_name in per_device_stats:
        bytes_read_delta = per_device_stats[device_name].kb_read - last_all_devices_stats.kb_read
        bytes_writ_delta = per_device_stats[device_name].kb_writ - last_all_devices_stats.kb_writ

    # There's a bug that sometimes the delta is negative messing up the average.
    bytes_read_delta = max(0, bytes_read_delta)
    bytes_read_delta = max(0, bytes_read_delta)

    aggr.mb_read_ps = bytes_read_delta * 512.0 / 1e6 / time_delta
    aggr.mb_writ_ps = bytes_writ_delta * 512.0 / 1e6 / time_delta

    io_read_delta: int = 0
    io_writ_delta: int = 0

    # Devices like AWS EBS or USB drives can get dynamically detached so validate key:
    if device_name in per_device_stats:
        io_read_delta = per_device_stats[device_name].io_read - last_all_devices_stats.io_read
        io_writ_delta = per_device_stats[device_name].io_writ - last_all_devices_stats.io_writ

    # There's a bug that sometimes the delta is negative messing up the average.
    io_read_delta = max(0, io_read_delta)
    io_writ_delta = max(0, io_writ_delta)

    aggr.io_read_ps = io_read_delta / time_delta
    aggr.io_writ_ps = io_writ_delta / time_delta

    util_delta: int = per_device_stats[device_name].io_util - last_all_devices_stats.io_util
    util_delta = max(0, util_delta)
    aggr.io_util = 100 * util_delta / (time_delta * 1000.0)
    aggr.io_util = min(100, aggr.io_util)

    return aggr


# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "DiskMetrics",
]
